#include "tts_tool.h"
#include <wchar.h>
#include <objbase.h>
#include <initguid.h>

/* 
*****************************************************
  Name:
    tts_tool.c

  Description:
      Functions for using SAPI4 compatible Text to Speech Synthesizers
  
  License:
      Copyright (c) 2002, Martin Goebbel
      All rights reserved.
      Redistribution and use in source and binary forms, with or without
      modification, are permitted provided that the following conditions
      are met:
        Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer. 

        The names of its contributors may not be used to endorse or 
        promote products derived from this software without specific 
        prior written permission. 
      
      THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS 
      IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED 
      TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
      PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE CONTRIBUTORS 
      BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
      CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
      SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
      INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
      CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
      ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
      THE POSSIBILITY OF SUCH DAMAGE.

  History:
    Date        By       Comment
    23.11.2001  martin   Initial Version 

*****************************************************
*/


#ifndef _MBCS
#ifndef _UNICODE

/* Either _MBCS or _UNICODE should be defined, _MBCS assumed */
#define _MBCS

#define _MBCS
#endif
#endif

typedef unsigned __int64 QWORD, *PQWORD;

typedef struct SData
{
  PVOID    pData;
  DWORD    dwSize;
} SDATA, *PSDATA;

typedef enum
{
  CHARSET_TEXT           = 0,
  CHARSET_IPAPHONETIC    = 1,
  CHARSET_ENGINEPHONETIC = 2
} VOICECHARSET;

#define CINTERFACE

DEFINE_GUID(CLSID_MMAudioDest, 0xcb96b400, 0xc743, 0x11cd, 0x80, 0xe5, 0x0, 0xaa, 0x0, 0x3e, 0x4b, 0x50);
DEFINE_GUID(IID_IAudioMultiMediaDevice, 0xb68ad320, 0xc743, 0x11cd, 0x80, 0xe5, 0x0, 0xaa, 0x0, 0x3e, 0x4b, 0x50);

DEFINE_GUID(CLSID_TTSEnumerator, 0xd67c0280, 0xc743, 0x11cd, 0x80, 0xe5, 0x0, 0xaa, 0x0, 0x3e, 0x4b, 0x50);
DEFINE_GUID(IID_ITTSEnumW, 0x6B837B20L, 0x4A47, 0x101B, 0x93, 0x1A, 0x00, 0xAA, 0x00, 0x47, 0xBA, 0x4F);
DEFINE_GUID(IID_ITTSCentralW, 0x28016060L, 0x4A47, 0x101B, 0x93, 0x1A, 0x00, 0xAA, 0x00, 0x47, 0xBA, 0x4F);
DEFINE_GUID(IID_ITTSNotifySinkW, 0xC0FA8F40L, 0x4A46, 0x101B, 0x93, 0x1A, 0x00, 0xAA, 0x00, 0x47, 0xBA, 0x4F);
DEFINE_GUID(IID_ITTSBufNotifySink, 0xe4963d40, 0xc743, 0x11cd, 0x80, 0xe5, 0x0, 0xaa, 0x0, 0x3e, 0x4b, 0x50);
DEFINE_GUID(IID_ITTSAttributesW, 0x1287A280L, 0x4A47, 0x101B, 0x93, 0x1A, 0x00, 0xAA, 0x00, 0x47, 0xBA, 0x4F);


#undef   INTERFACE
#define  INTERFACE   IAudioMultiMediaDevice

DECLARE_INTERFACE_ (IAudioMultiMediaDevice, IUnknown)
{
   // IUnknown members
   STDMETHOD (QueryInterface) (THIS_ REFIID riid, LPVOID FAR* ppvObj) PURE;
   STDMETHOD_(ULONG,AddRef)   (THIS) PURE;
   STDMETHOD_(ULONG,Release)  (THIS) PURE;

   // IAudioMultiMediaDevice members
   STDMETHOD (CustomMessage)  (THIS_ UINT, SDATA) PURE;
   STDMETHOD (DeviceNumGet)   (THIS_ DWORD*) PURE;
   STDMETHOD (DeviceNumSet)   (THIS_ DWORD) PURE;
};
typedef IAudioMultiMediaDevice FAR * PIAUDIOMULTIMEDIADEVICE;

#undef   INTERFACE
#define  INTERFACE   ITTSCentralW
DECLARE_INTERFACE_ (ITTSCentralW, IUnknown)
{
   // IUnknown members
   STDMETHOD (QueryInterface) (THIS_ REFIID riid, LPVOID FAR* ppvObj) PURE;
   STDMETHOD_(ULONG,AddRef)   (THIS) PURE;
   STDMETHOD_(ULONG,Release)  (THIS) PURE;

   // ITTSCentralW members
   STDMETHOD (Inject)         (THIS_ PCWSTR) PURE;
   STDMETHOD (ModeGet)        (THIS_ PTTSMODEINFOW) PURE;
   STDMETHOD (Phoneme)        (THIS_ VOICECHARSET, DWORD, SDATA, PSDATA) PURE;
   STDMETHOD (PosnGet)        (THIS_ PQWORD) PURE;
   STDMETHOD (TextData)       (THIS_ VOICECHARSET, DWORD, SDATA, PVOID, IID) PURE;
   STDMETHOD (ToFileTime)     (THIS_ PQWORD, FILETIME *) PURE;
   STDMETHOD (AudioPause)     (THIS) PURE;
   STDMETHOD (AudioResume)    (THIS) PURE;
   STDMETHOD (AudioReset)     (THIS) PURE;
   STDMETHOD (Register)       (THIS_ PVOID, IID, DWORD*) PURE;
   STDMETHOD (UnRegister)     (THIS_ DWORD) PURE;
};
typedef ITTSCentralW FAR * PITTSCENTRALW;

#undef   INTERFACE
#define  INTERFACE   ITTSEnumW
DECLARE_INTERFACE_ (ITTSEnumW, IUnknown)
{
   // IUnknown members
   STDMETHOD (QueryInterface) (THIS_ REFIID riid, LPVOID FAR* ppvObj) PURE;
   STDMETHOD_(ULONG,AddRef)   (THIS) PURE;
   STDMETHOD_(ULONG,Release)  (THIS) PURE;

   // ITTSEnumW members
   STDMETHOD (Next)           (THIS_ ULONG, PTTSMODEINFOW, ULONG *) PURE;
   STDMETHOD (Skip)           (THIS_ ULONG) PURE;
   STDMETHOD (Reset)          (THIS) PURE;
   STDMETHOD (Clone)          (THIS_ ITTSEnumW * FAR *) PURE;
   STDMETHOD (Select)         (THIS_ GUID, PITTSCENTRALW *, LPUNKNOWN) PURE;
};

#undef   INTERFACE
#define  INTERFACE   ITTSNotifySinkW
DECLARE_INTERFACE_ (ITTSNotifySinkW, IUnknown)
{
   STDMETHOD (QueryInterface) (THIS_ REFIID riid, LPVOID FAR* ppvObj) PURE;
   STDMETHOD_(ULONG,AddRef)   (THIS) PURE;
   STDMETHOD_(ULONG,Release)  (THIS) PURE;
   STDMETHOD (AttribChanged)  (THIS_ DWORD) PURE;
   STDMETHOD (AudioStart)     (THIS_ QWORD) PURE;
   STDMETHOD (AudioStop)      (THIS_ QWORD) PURE;
   STDMETHOD (Visual)         (THIS_ QWORD, WCHAR, WCHAR, DWORD, LPVOID) PURE;
};
typedef ITTSNotifySinkW FAR * PITTSNOTIFYSINKW;

#undef   INTERFACE
#define  INTERFACE   ITTSBufNotifySink
DECLARE_INTERFACE_ (ITTSBufNotifySink, IUnknown)
{
   STDMETHOD(QueryInterface)  (THIS_ REFIID riid, LPVOID FAR* ppvObj) PURE;
   STDMETHOD_(ULONG,AddRef)   (THIS) PURE;
   STDMETHOD_(ULONG,Release)  (THIS) PURE;
   STDMETHOD (TextDataDone)   (THIS_ QWORD, DWORD) PURE;
   STDMETHOD (TextDataStarted)(THIS_ QWORD) PURE;
   STDMETHOD (BookMark)       (THIS_ QWORD, DWORD) PURE;
   STDMETHOD (WordPosition)   (THIS_ QWORD, DWORD) PURE;
};
typedef ITTSBufNotifySink FAR * PITTSBUFNOTIFYSINK;

#undef   INTERFACE
#define  INTERFACE   ITTSAttributesW
DECLARE_INTERFACE_ (ITTSAttributesW, IUnknown)
{
   STDMETHOD(QueryInterface)  (THIS_ REFIID riid, LPVOID FAR* ppvObj) PURE;
   STDMETHOD_(ULONG,AddRef)   (THIS) PURE;
   STDMETHOD_(ULONG,Release)  (THIS) PURE;
   STDMETHOD (PitchGet)       (THIS_ WORD *) PURE;
   STDMETHOD (PitchSet)       (THIS_ WORD) PURE;
   STDMETHOD (RealTimeGet)    (THIS_ DWORD *) PURE;
   STDMETHOD (RealTimeSet)    (THIS_ DWORD) PURE;
   STDMETHOD (SpeedGet)       (THIS_ DWORD *) PURE;
   STDMETHOD (SpeedSet)       (THIS_ DWORD) PURE;
   STDMETHOD (VolumeGet)      (THIS_ DWORD *) PURE;
   STDMETHOD (VolumeSet)      (THIS_ DWORD) PURE;
};
typedef ITTSAttributesW FAR * PITTSATTRIBUTESW;


typedef struct _NotifySink
{
  ITTSNotifySinkW iface;

} NotifySink;

static HRESULT STDMETHODCALLTYPE NotifySink_QueryInterface ( ITTSNotifySinkW *This, REFIID riid, LPVOID *ppvObj )
{
	*ppvObj = NULL;

	if (IsEqualIID (riid, &IID_IUnknown) || IsEqualIID(riid, &IID_ITTSNotifySinkW))
	{
		*ppvObj = (LPVOID)This;
		return S_OK;
	}
	return E_NOINTERFACE;
}

static ULONG STDMETHODCALLTYPE NotifySink_AddRef ( ITTSNotifySinkW *This )
{
  return 1;
}

static ULONG STDMETHODCALLTYPE NotifySink_Release ( ITTSNotifySinkW *This )
{
  return 1;
}

static HRESULT STDMETHODCALLTYPE NotifySink_AttribChanged ( ITTSNotifySinkW *This, DWORD dwAttribute )
{
	return S_OK;
}

static HRESULT STDMETHODCALLTYPE NotifySink_AudioStart (ITTSNotifySinkW *This, QWORD qTimeStamp )
{
	return S_OK;
}

static HRESULT STDMETHODCALLTYPE NotifySink_AudioStop (ITTSNotifySinkW *This, QWORD qTimeStamp )
{
	return S_OK;
}

static HRESULT STDMETHODCALLTYPE NotifySink_Visual (ITTSNotifySinkW *This, QWORD qTimeStamp, WCHAR cIPAPhoneme, WCHAR cEnginePhoneme, DWORD dwHints, LPVOID pTTSMouth )
{
	return S_OK;
}

static ITTSNotifySinkWVtbl NotifySinkVtbl =
{
  NotifySink_QueryInterface,
  NotifySink_AddRef,
  NotifySink_Release,
  NotifySink_AttribChanged,
  NotifySink_AudioStart,
  NotifySink_AudioStop,
  NotifySink_Visual
};


typedef struct _BufNotifySink
{
  ITTSBufNotifySink iface;
  TTSCALLBACK       callback;
} BufNotifySink;

static HRESULT STDMETHODCALLTYPE BufNotifySink_QueryInterface ( ITTSBufNotifySink *This, REFIID riid, LPVOID *ppvObj )
{
	*ppvObj = NULL;

	if (IsEqualIID (riid, &IID_IUnknown) || IsEqualIID(riid, &IID_ITTSBufNotifySink))
	{
		*ppvObj = (LPVOID)This;
		return S_OK;
	}
	return E_NOINTERFACE;
}

static ULONG STDMETHODCALLTYPE BufNotifySink_AddRef ( ITTSBufNotifySink *This )
{
  return 1;
}

static ULONG STDMETHODCALLTYPE BufNotifySink_Release ( ITTSBufNotifySink *This )
{
  return 1;
}

static HRESULT STDMETHODCALLTYPE BufNotifySink_TextDataDone ( ITTSBufNotifySink *This, QWORD qTimeStamp, DWORD dwFlags )
{
  BufNotifySink *pInstance;

  if ( (pInstance = (BufNotifySink *)This) != NULL )
  {
    if ( pInstance->callback != NULL )
    {
      pInstance->callback(TTS_EVENT_TEXTDONE);
    }
  }

  return S_OK;
}

static HRESULT STDMETHODCALLTYPE BufNotifySink_TextDataStarted ( ITTSBufNotifySink *This, QWORD qTimeStamp )
{
  BufNotifySink *pInstance;

  if ( (pInstance = (BufNotifySink *)This) != NULL )
  {
    if ( pInstance->callback != NULL )
    {
      pInstance->callback(TTS_EVENT_TEXTSTART);
    }
  }

	return S_OK;
}

static HRESULT STDMETHODCALLTYPE BufNotifySink_BookMark ( ITTSBufNotifySink *This, QWORD qTimeStamp, DWORD dwMarkNum )
{
	return S_OK;
}

static HRESULT STDMETHODCALLTYPE BufNotifySink_WordPosition ( ITTSBufNotifySink *This, QWORD qTimeStamp, DWORD dwByteOffset )
{
	return S_OK;
}

static ITTSBufNotifySinkVtbl BufNotifySinkVtbl =
{
  BufNotifySink_QueryInterface,
  BufNotifySink_AddRef,
  BufNotifySink_Release,
  BufNotifySink_TextDataDone,
  BufNotifySink_TextDataStarted,
  BufNotifySink_BookMark,
  BufNotifySink_WordPosition
};

typedef struct _TtsSynth
{
  PITTSCENTRALW     pTts;
  PITTSATTRIBUTESW  pAttr;
  DWORD             dwReg;
  DWORD             dwMinSpeed, dwMaxSpeed;
  WORD              wMinPitch, wMaxPitch;
  DWORD             dwMinVol, dwMaxVol;
  NotifySink        notify;
  BufNotifySink     bufnotify;
} TtsSynth;

TTSENUM tts_EnumCreate ( void )
{
  HRESULT hr;
  ITTSEnumW *pEnum;

  hr = CoCreateInstance(&CLSID_TTSEnumerator, NULL, CLSCTX_ALL, &IID_ITTSEnumW, (LPVOID *)&pEnum);
  if ( SUCCEEDED(hr) )
  {
    return (TTSENUM) pEnum;
  }

 return NULL;
}

void tts_EnumDelete ( TTSENUM ttsEnum )
{
  ITTSEnumW *pEnum;

  if ( (pEnum = (ITTSEnumW *)ttsEnum) != NULL )
  {
    pEnum->lpVtbl->Release(pEnum);
  }
}

BOOL tts_EnumReset ( TTSENUM ttsEnum )
{
  ITTSEnumW *pEnum;

  if ( (pEnum = (ITTSEnumW *)ttsEnum) != NULL )
  {
    return ( SUCCEEDED(pEnum->lpVtbl->Reset(pEnum)) );
  }

  return FALSE;
}

BOOL tts_EnumNext ( TTSENUM ttsEnum, PTTSMODEINFOW pTtsModeInfo )
{
  ITTSEnumW *pEnum;
  UINT nCount;

  if ( ((pEnum = (ITTSEnumW *)ttsEnum) != NULL) && (pTtsModeInfo != NULL) )
  {
    nCount = 0;

    if ( SUCCEEDED(pEnum->lpVtbl->Next(pEnum, 1, pTtsModeInfo, &nCount)) && (nCount == 1) )
    {
      return TRUE;
    }
  }

  return FALSE;
}

TTS tts_Create ( PTTSMODEINFOW pTtsModeInfo, TTSCALLBACK lpfnCallback )
{
  TtsSynth *pSynth;
  HRESULT hr;
  ITTSEnumW *pEnum;
  PIAUDIOMULTIMEDIADEVICE pIMMDevice;
  IUnknown *pUnkAudio;
  DWORD dwCur;
  WORD  wCur;
  BOOL bOk;

  if ( pTtsModeInfo != NULL )
  {
    pSynth = (TtsSynth *)malloc(sizeof(TtsSynth));
    if ( pSynth == NULL )
    {
      return NULL;
    }
    pSynth->pTts = NULL;
    pSynth->notify.iface.lpVtbl = &NotifySinkVtbl;
    pSynth->bufnotify.iface.lpVtbl = &BufNotifySinkVtbl;
    pSynth->bufnotify.callback = lpfnCallback;

    hr = CoCreateInstance(&CLSID_TTSEnumerator, NULL, CLSCTX_ALL, &IID_ITTSEnumW, (LPVOID *)&pEnum);
    if ( FAILED(hr) )
    {
      free(pSynth);
      return NULL;
    }

    hr = CoCreateInstance(&CLSID_MMAudioDest, NULL, CLSCTX_ALL, &IID_IAudioMultiMediaDevice, (LPVOID *)&pUnkAudio);
    if ( FAILED(hr) )
    {
      free(pSynth);
      pEnum->lpVtbl->Release(pEnum);
      return FALSE;
    }

    if ( FAILED(pUnkAudio->lpVtbl->QueryInterface(pUnkAudio, &IID_IAudioMultiMediaDevice, (void **)&pIMMDevice)) )
    {
      free(pSynth);
      pEnum->lpVtbl->Release(pEnum);
      pUnkAudio->lpVtbl->Release(pUnkAudio);
      return FALSE;
    }
    pIMMDevice->lpVtbl->DeviceNumSet(pIMMDevice, (UINT)-1);
    pIMMDevice->lpVtbl->Release(pIMMDevice);


    hr = pEnum->lpVtbl->Select(pEnum, pTtsModeInfo->gModeID, &pSynth->pTts, pUnkAudio);

    pEnum->lpVtbl->Release(pEnum);
    pUnkAudio->lpVtbl->Release(pUnkAudio);

    if ( SUCCEEDED(hr) )
    {
      if ( FAILED(pSynth->pTts->lpVtbl->Register(pSynth->pTts, (void *)&pSynth->notify, IID_ITTSNotifySinkW, &pSynth->dwReg)) )
      {
        pSynth->pTts->lpVtbl->Release(pSynth->pTts);
        free(pSynth);
        return NULL;
      }

      if ( FAILED(pSynth->pTts->lpVtbl->QueryInterface(pSynth->pTts, &IID_ITTSAttributesW, (PVOID*)&pSynth->pAttr)) )
      {
        pSynth->pAttr = NULL;
      }
      else
      {
        bOk  = ( SUCCEEDED(pSynth->pAttr->lpVtbl->VolumeGet(pSynth->pAttr, &dwCur)) );
        bOk &= ( SUCCEEDED(pSynth->pAttr->lpVtbl->VolumeSet(pSynth->pAttr, 0)) );
        bOk &= ( SUCCEEDED(pSynth->pAttr->lpVtbl->VolumeGet(pSynth->pAttr, &pSynth->dwMinVol)) );
        bOk &= ( SUCCEEDED(pSynth->pAttr->lpVtbl->VolumeSet(pSynth->pAttr, 0xffffffff)) );
        bOk &= ( SUCCEEDED(pSynth->pAttr->lpVtbl->VolumeGet(pSynth->pAttr, &pSynth->dwMaxVol)) );
        bOk &= ( SUCCEEDED(pSynth->pAttr->lpVtbl->VolumeSet(pSynth->pAttr, dwCur)) );
        if ( !bOk )
        {
          pSynth->dwMaxVol = 0;
          pSynth->dwMinVol = 0;
        }

        bOk  = ( SUCCEEDED(pSynth->pAttr->lpVtbl->SpeedGet(pSynth->pAttr, &dwCur)) );
        bOk &= ( SUCCEEDED(pSynth->pAttr->lpVtbl->SpeedSet(pSynth->pAttr, 0)) );
        bOk &= ( SUCCEEDED(pSynth->pAttr->lpVtbl->SpeedGet(pSynth->pAttr, &pSynth->dwMinSpeed)) );
        bOk &= ( SUCCEEDED(pSynth->pAttr->lpVtbl->SpeedSet(pSynth->pAttr, 0xffffffff)) );
        bOk &= ( SUCCEEDED(pSynth->pAttr->lpVtbl->SpeedGet(pSynth->pAttr, &pSynth->dwMaxSpeed)) );
        bOk &= ( SUCCEEDED(pSynth->pAttr->lpVtbl->SpeedSet(pSynth->pAttr, dwCur)) );
        if ( !bOk )
        {
          pSynth->dwMaxSpeed = 0;
          pSynth->dwMinSpeed = 0;
        }

        bOk  = ( SUCCEEDED(pSynth->pAttr->lpVtbl->PitchGet(pSynth->pAttr, &wCur)) );
        bOk &= ( SUCCEEDED(pSynth->pAttr->lpVtbl->PitchSet(pSynth->pAttr, 0)) );
        bOk &= ( SUCCEEDED(pSynth->pAttr->lpVtbl->PitchGet(pSynth->pAttr, &pSynth->wMinPitch)) );
        bOk &= ( SUCCEEDED(pSynth->pAttr->lpVtbl->PitchSet(pSynth->pAttr, 0xffff)) );
        bOk &= ( SUCCEEDED(pSynth->pAttr->lpVtbl->PitchGet(pSynth->pAttr, &pSynth->wMaxPitch)) );
        bOk &= ( SUCCEEDED(pSynth->pAttr->lpVtbl->PitchSet(pSynth->pAttr, wCur)) );
        if ( !bOk )
        {
          pSynth->wMaxPitch = 0;
          pSynth->wMinPitch = 0;
        }
      }

      return (TTS)pSynth;
    }

    free(pSynth);
  }

  return NULL;
}

void tts_Delete ( TTS Tts )
{
  TtsSynth *pSynth;

  if ( ((pSynth = (TtsSynth *)Tts) != NULL) && (pSynth->pTts != NULL) )
  {
    pSynth->pTts->lpVtbl->AudioReset(pSynth->pTts);
    pSynth->pTts->lpVtbl->UnRegister(pSynth->pTts, pSynth->dwReg);

    if ( pSynth->pAttr != NULL )
    {
      pSynth->pAttr->lpVtbl->Release(pSynth->pAttr);
    }

    pSynth->pTts->lpVtbl->Release(pSynth->pTts);

    free(pSynth);
  }
}

BOOL tts_Stop ( TTS Tts )
{
  TtsSynth *pSynth;

  if ( ((pSynth = (TtsSynth *)Tts) != NULL) && (pSynth->pTts != NULL) )
  {
    return SUCCEEDED(pSynth->pTts->lpVtbl->AudioReset(pSynth->pTts));
  }

  return FALSE;
}

BOOL tts_Pause ( TTS Tts )
{
  TtsSynth *pSynth;

  if ( ((pSynth = (TtsSynth *)Tts) != NULL) && (pSynth->pTts != NULL) )
  {
    return SUCCEEDED(pSynth->pTts->lpVtbl->AudioPause(pSynth->pTts));
  }

  return FALSE;
}

BOOL tts_Resume ( TTS Tts )
{
  TtsSynth *pSynth;

  if ( ((pSynth = (TtsSynth *)Tts) != NULL) && (pSynth->pTts != NULL) )
  {
    return SUCCEEDED(pSynth->pTts->lpVtbl->AudioResume(pSynth->pTts));
  }

  return FALSE;
}

BOOL tts_TextOut ( TTS Tts, LPCTSTR lpszText )
{
  TtsSynth *pSynth;
  HRESULT hr;
  SDATA text;
  int nSize;

  if ( ((pSynth = (TtsSynth *)Tts) != NULL) && (pSynth->pTts != NULL) && (lpszText != NULL) )
  {
    if ( *lpszText == _T(0) )
    {
      return TRUE;
    }

#ifndef _MBCS
    text.dwSize = (wcslen(lpszText) + 1) * sizeof(WCHAR);
    text.pData  = (PVOID)lpszText;
#else
    nSize = MultiByteToWideChar(CP_ACP, 0, lpszText, -1, NULL, 0);
    if ( nSize <= 0 )
    {
      return FALSE;
    }

		text.dwSize = nSize * sizeof(WCHAR);
    text.pData = malloc(text.dwSize);
    if ( text.pData == NULL )
    {
      return FALSE;
    }

    if ( MultiByteToWideChar(CP_ACP, 0, lpszText, -1, (LPWSTR)text.pData, nSize) <= 0 )
    {
      free(text.pData);
      return FALSE;
    }
    text.dwSize = (wcslen((LPWSTR)text.pData) + 1) * sizeof(WCHAR);
#endif

    hr = pSynth->pTts->lpVtbl->TextData(pSynth->pTts, CHARSET_TEXT, 0, text, (void *)&pSynth->bufnotify, IID_ITTSBufNotifySink);

#ifdef _MBCS
    free(text.pData);
#endif

    return SUCCEEDED(hr);
  }

  return FALSE;
}

BOOL tts_SetVolume ( TTS Tts, UINT nVolPercent )
{
  TtsSynth *pSynth;
  DWORD dwNewVol;

  if ( ((pSynth = (TtsSynth *)Tts) != NULL) && (pSynth->pAttr != NULL) )
  {
    if ( (pSynth->dwMaxVol == 0) && (pSynth->dwMinVol == 0) )
    {
      return FALSE;
    }

    if ( nVolPercent == 0 )
    {
      dwNewVol = pSynth->dwMinVol;
    }
    else if ( nVolPercent >= 100 )
    {
      dwNewVol = pSynth->dwMaxVol;
    }
    else
    {
      dwNewVol = pSynth->dwMinVol + (DWORD)((((QWORD)pSynth->dwMaxVol - (QWORD)pSynth->dwMinVol) * (QWORD)nVolPercent) / (QWORD)100);
    }

    return SUCCEEDED(pSynth->pAttr->lpVtbl->VolumeSet(pSynth->pAttr, dwNewVol));
  }

  return FALSE;
}

BOOL tts_GetVolume ( TTS Tts, UINT *pVolPercent )
{
  TtsSynth *pSynth;
  DWORD dwCurVol;

  if ( ((pSynth = (TtsSynth *)Tts) != NULL) && (pSynth->pAttr != NULL) && (pVolPercent != NULL) )
  {
    if ( (pSynth->dwMaxVol == 0) && (pSynth->dwMinVol == 0) )
    {
      return FALSE;
    }

    if ( SUCCEEDED(pSynth->pAttr->lpVtbl->VolumeGet(pSynth->pAttr, &dwCurVol)) )
    {
      *pVolPercent = (UINT)(((QWORD)dwCurVol * (QWORD)100) / ((QWORD)pSynth->dwMaxVol - (QWORD)pSynth->dwMinVol));
      return TRUE;
    }
  }

  return FALSE;
}

BOOL tts_SetSpeed ( TTS Tts, UINT nSpeedPercent )
{
  TtsSynth *pSynth;
  DWORD dwNewSpeed;

  if ( ((pSynth = (TtsSynth *)Tts) != NULL) && (pSynth->pAttr != NULL) )
  {
    if ( (pSynth->dwMaxSpeed == 0) && (pSynth->dwMinSpeed == 0) )
    {
      return FALSE;
    }

    if ( nSpeedPercent == 0 )
    {
      dwNewSpeed = pSynth->dwMinSpeed;
    }
    else if ( nSpeedPercent >= 100 )
    {
      dwNewSpeed = pSynth->dwMaxSpeed;
    }
    else
    {
      dwNewSpeed = pSynth->dwMinSpeed + (DWORD)((((QWORD)pSynth->dwMaxSpeed - (QWORD)pSynth->dwMinSpeed) * (QWORD)nSpeedPercent) / (QWORD)100);
    }

    return SUCCEEDED(pSynth->pAttr->lpVtbl->SpeedSet(pSynth->pAttr, dwNewSpeed));
  }

  return FALSE;
}

BOOL tts_GetSpeed ( TTS Tts, UINT *pSpeedPercent )
{
  TtsSynth *pSynth;
  DWORD dwCurSpeed;

  if ( ((pSynth = (TtsSynth *)Tts) != NULL) && (pSynth->pAttr != NULL) && (pSpeedPercent != NULL) )
  {
    if ( (pSynth->dwMaxSpeed == 0) && (pSynth->dwMinSpeed == 0) )
    {
      return FALSE;
    }

    if ( SUCCEEDED(pSynth->pAttr->lpVtbl->SpeedGet(pSynth->pAttr, &dwCurSpeed)) )
    {
      *pSpeedPercent = (UINT)(((QWORD)dwCurSpeed * (QWORD)100) / ((QWORD)pSynth->dwMaxSpeed - (QWORD)pSynth->dwMinSpeed));
      return TRUE;
    }
  }

  return FALSE;
}

BOOL tts_SetPitch ( TTS Tts, UINT nPitchPercent )
{
  TtsSynth *pSynth;
  WORD wNewPitch;

  if ( ((pSynth = (TtsSynth *)Tts) != NULL) && (pSynth->pAttr != NULL) )
  {
    if ( (pSynth->wMaxPitch == 0) && (pSynth->wMinPitch == 0) )
    {
      return FALSE;
    }

    if ( nPitchPercent == 0 )
    {
      wNewPitch = pSynth->wMinPitch;
    }
    else if ( nPitchPercent >= 100 )
    {
      wNewPitch = pSynth->wMaxPitch;
    }
    else
    {
      wNewPitch = pSynth->wMinPitch + (WORD)((((DWORD)pSynth->wMaxPitch - (DWORD)pSynth->wMinPitch) * (DWORD)nPitchPercent) / (DWORD)100);
    }

    return SUCCEEDED(pSynth->pAttr->lpVtbl->PitchSet(pSynth->pAttr, wNewPitch));
  }

  return FALSE;
}

BOOL tts_GetPitch ( TTS Tts, UINT *pPitchPercent )
{
  TtsSynth *pSynth;
  WORD wCurPitch;

  if ( ((pSynth = (TtsSynth *)Tts) != NULL) && (pSynth->pAttr != NULL) && (pPitchPercent != NULL) )
  {
    if ( (pSynth->wMaxPitch == 0) && (pSynth->wMinPitch == 0) )
    {
      return FALSE;
    }

    if ( SUCCEEDED(pSynth->pAttr->lpVtbl->PitchGet(pSynth->pAttr, &wCurPitch)) )
    {
      *pPitchPercent = (UINT)(((DWORD)wCurPitch * (DWORD)100) / ((DWORD)pSynth->wMaxPitch - (QWORD)pSynth->wMinPitch));
      return TRUE;
    }
  }

  return FALSE;
}

